from contextlib import suppress
from datetime import datetime, timedelta
from hashlib import md5
from os import environ, mkdir, rename
from os.path import dirname, exists, join
from threading import Lock

from tomlkit import dumps as toml_dump, parse as toml_parse

from . import operations
from .bundle import Bundle
from .concurrency import WorkerPool
from .deps import find_item, ItemDependencyLoop
from .exceptions import (
    BundleError,
    GracefulApplyException,
    ItemSkipped,
    NodeLockedException,
    NoSuchBundle,
    NoSuchGroup,
    RemoteException,
    TransportException,
    RepositoryError,
    SkipNode,
)
from .group import GROUP_ATTR_DEFAULTS, GROUP_ATTR_TYPES, GROUP_ATTR_TYPES_ENFORCED
from .itemqueue import ItemQueue
from .items import Item
from .lock import NodeLock, softlock_add
from .metadata import hash_metadata
from .utils import (
    cached_property,
    cached_property_set,
    error_context,
    get_file_contents,
    names,
)
from .utils.dicts import (
    dict_to_text,
    dict_to_toml,
    diff_dict,
    hash_state_dict,
    set_key_at_path,
    normalize_dict,
    validate_dict,
    COLLECTION_OF_STRINGS,
)
from .utils.magic_strings import convert_magic_strings
from .utils.text import (
    blue,
    bold,
    cyan,
    force_text,
    format_duration,
    green,
    mark_for_translation as _,
    prefix_lines,
    randstr,
    red,
    toml_clean,
    validate_name,
    yellow,
)
from .utils.ui import io



NODE_ATTR_TYPES = GROUP_ATTR_TYPES.copy()
NODE_ATTR_TYPES['groups'] = COLLECTION_OF_STRINGS
NODE_ATTR_TYPES['hostname'] = str
NODE_ATTRS = sorted(list(GROUP_ATTR_DEFAULTS) + ['bundles', 'file_path', 'groups', 'hostname'])


class ApplyResult:
    """
    Holds information about an apply run for a node.
    """
    def __init__(self, node, item_results):
        self.node_name = node.name
        self.correct = 0
        self.fixed = 0
        self.skipped = 0
        self.failed = 0
        self.total = 0

        for item_id, result, duration in item_results:
            self.total += 1
            if result == Item.STATUS_ACTION_SUCCEEDED:
                self.correct += 1
            elif result == Item.STATUS_OK:
                self.correct += 1
            elif result == Item.STATUS_FIXED:
                self.fixed += 1
            elif result == Item.STATUS_SKIPPED:
                self.skipped += 1
            elif result == Item.STATUS_FAILED:
                self.failed += 1
            else:
                raise RuntimeError(_(
                    "can't make sense of results for {} on {}: {}"
                ).format(item_id, self.node_name, result))

        self.start = None
        self.end = None

    def __lt__(self, other):
        return self.node_name < other.node_name

    @property
    def duration(self):
        return self.end - self.start


def format_node_result(result):
    output = []
    output.append(("{count} OK").format(count=result.correct))

    if result.fixed:
        output.append(green(_("{count} fixed").format(count=result.fixed)))
    else:
        output.append(_("{count} fixed").format(count=result.fixed))

    if result.skipped:
        output.append(yellow(_("{count} skipped").format(count=result.skipped)))
    else:
        output.append(_("{count} skipped").format(count=result.skipped))

    if result.failed:
        output.append(red(_("{count} failed").format(count=result.failed)))
    else:
        output.append(_("{count} failed").format(count=result.failed))

    return ", ".join(output)


def handle_apply_result(
    node,
    item,
    status_code,
    interactive=False,
    details=None,
    show_diff=True,
    created=None,
    deleted=None,
    show_skipped_items=True,
):
    if status_code == Item.STATUS_SKIPPED and details in (
        Item.SKIP_REASON_NO_TRIGGER,
        Item.SKIP_REASON_UNLESS,
    ):
        # skipped for "unless" or "not triggered", don't output those
        return

    if status_code == Item.STATUS_SKIPPED and not show_skipped_items:
        return

    formatted_result = format_item_result(
        status_code,
        node.name,
        item.bundle.name,
        item.id,
        interactive=interactive,
        details=details,
        show_diff=show_diff,
        created=created,
        deleted=deleted,
    )
    if formatted_result is not None:
        if status_code == Item.STATUS_FAILED:
            io.stderr(formatted_result)
            if item._command_results:
                io.stderr(format_item_command_results(item._command_results))
                # free up memory
                del item._command_results
        else:
            io.stdout(formatted_result)


def apply_items(
    node,
    autoskip_selector="",
    autoonly_selector="",
    my_soft_locks=(),
    other_peoples_soft_locks=(),
    workers=1,
    interactive=False,
    show_diff=True,
    show_skipped_items=True,
):
    item_queue = ItemQueue(node)
    # the item queue might contain new generated items (canned actions)
    # adjust progress total accordingly
    extra_items = len(item_queue.all_items) - len(node.items)
    io.progress_increase_total(increment=extra_items)

    results = []

    def tasks_available():
        # Some item types are not allowed to run at the same time as
        # some other item types. items_without_deps_runnable handles
        # these cases.
        return bool(item_queue.items_without_deps_runnable())

    def next_task():
        item = item_queue.pop()
        return {
            'task_id': "{}:{}".format(node.name, item.id),
            'target': item.apply,
            'kwargs': {
                'autoskip_selector': autoskip_selector,
                'autoonly_selector': autoonly_selector,
                'my_soft_locks': my_soft_locks,
                'other_peoples_soft_locks': other_peoples_soft_locks,
                'interactive': interactive,
                'show_diff': show_diff,
            },
        }

    def handle_result(task_id, return_value, duration):
        item_id = task_id.split(":", 1)[1]
        item = find_item(item_id, item_queue.pending_items)

        status_code, details, created, deleted = return_value

        if status_code == Item.STATUS_FAILED:
            for skipped_item in item_queue.item_failed(item):
                handle_apply_result(
                    node,
                    skipped_item,
                    Item.STATUS_SKIPPED,
                    interactive=interactive,
                    details=Item.SKIP_REASON_DEP_FAILED,
                    show_skipped_items=show_skipped_items,
                )
                results.append((skipped_item.id, Item.STATUS_SKIPPED, timedelta(0)))
        elif status_code in (Item.STATUS_FIXED, Item.STATUS_ACTION_SUCCEEDED):
            item_queue.item_fixed(item)
        elif status_code == Item.STATUS_OK:
            item_queue.item_ok(item)
        elif status_code == Item.STATUS_SKIPPED:
            for skipped_item in item_queue.item_skipped(item):
                skip_reason = Item.SKIP_REASON_DEP_SKIPPED
                for lock in other_peoples_soft_locks:
                    for selector in lock['items']:
                        if skipped_item.covered_by_autoskip_selector(selector):
                            skip_reason = Item.SKIP_REASON_SOFTLOCK
                            break
                handle_apply_result(
                    node,
                    skipped_item,
                    Item.STATUS_SKIPPED,
                    interactive=interactive,
                    details=skip_reason,
                    show_skipped_items=show_skipped_items,
                )
                results.append((skipped_item.id, Item.STATUS_SKIPPED, timedelta(0)))
        else:
            raise AssertionError(_(
                "unknown item status returned for {item}: {status}".format(
                    item=item.id,
                    status=repr(status_code),
                ),
            ))

        handle_apply_result(
            node,
            item,
            status_code,
            interactive=interactive,
            details=details,
            show_diff=show_diff,
            created=created,
            deleted=deleted,
            show_skipped_items=show_skipped_items,
        )
        io.progress_advance()
        results.append((item.id, status_code, duration))

    def handle_exception(task_id, exc, traceback):
        item_id = task_id.split(":", 1)[1]
        item = find_item(item_id, item_queue.pending_items)

        for skipped_item in item_queue.item_failed(item):
            handle_apply_result(
                node,
                skipped_item,
                Item.STATUS_SKIPPED,
                interactive=interactive,
                details=Item.SKIP_REASON_DEP_FAILED,
                show_skipped_items=show_skipped_items,
            )
            results.append((skipped_item.id, Item.STATUS_SKIPPED, timedelta(0)))

        handle_apply_result(
            node,
            item,
            Item.STATUS_FAILED,
            interactive=interactive,
            details=None,
            show_diff=False,
            show_skipped_items=show_skipped_items,
        )
        results.append((item.id, Item.STATUS_FAILED, timedelta(0)))
        output = prefix_lines(traceback, red("│ "))
        output += prefix_lines(repr(exc), red("│ "))
        output += red("╵")
        io.stderr(output)
        io.progress_advance()

    worker_pool = WorkerPool(
        tasks_available,
        next_task,
        handle_result=handle_result,
        handle_exception=handle_exception,
        pool_id="apply_{}".format(node.name),
        workers=workers,
    )
    worker_pool.run()

    # we have no items without deps left and none are processing
    # there must be a loop
    if item_queue.items_with_deps:
        raise ItemDependencyLoop(item_queue.items_with_deps)

    return results


def _flatten_group_hierarchy(groups):
    """
    Takes a list of groups and returns a list of group names ordered so
    that parent groups will appear before any of their subgroups.
    """
    # dict mapping groups to subgroups
    child_groups = {}
    for group in groups:
        child_groups[group.name] = list(names(group.subgroups))

    # dict mapping groups to parent groups
    parent_groups = {}
    for child_group in child_groups:
        parent_groups[child_group] = []
        for parent_group, subgroups in child_groups.items():
            if child_group in subgroups:
                parent_groups[child_group].append(parent_group)

    order = []

    while True:
        top_level_group = None
        for group, parents in parent_groups.items():
            if parents:
                continue
            else:
                top_level_group = group
                break
        if not top_level_group:
            if parent_groups:
                raise RuntimeError(
                    _("encountered subgroup loop that should have been detected")
                )
            else:
                break
        order.append(top_level_group)
        del parent_groups[top_level_group]
        for group in parent_groups:
            if top_level_group in parent_groups[group]:
                parent_groups[group].remove(top_level_group)

    return order


def format_item_command_results(results):
    output = ""

    for result in results:
        stdout = result['result'].stdout_text.strip()
        stderr = result['result'].stderr_text.strip()

        # show command
        output += "{b}".format(b=red('│'))
        output += "\n{b} {command} (return code: {code}{no_output})\n".format(
            b=red('├─'),
            command=bold(result['command']),
            code=bold(result['result'].return_code),
            no_output='' if stdout or stderr else '; no output'
        )

        # show output
        lines = ""
        if stdout or stderr:
            lines += "\n"
            if stdout:
                lines += stdout.strip()
            if stderr:
                lines += stderr.strip()

        output += prefix_lines(lines, red("│ "))

    output += red("╵")
    return output.lstrip('\n')


def format_item_result(
    result,
    node,
    bundle,
    item,
    interactive=False,
    details=None,
    show_diff=True,
    created=None,
    deleted=None,
):
    if created or deleted or details is None:
        details_text = ""
    elif result == Item.STATUS_SKIPPED:
        details_text = "({})".format(Item.SKIP_REASON_DESC[details])
    else:
        details_text = "({})".format(", ".join(sorted(details[2])))
    if result == Item.STATUS_FAILED:
        if created:
            status = red(_("failed to create"))
        elif deleted:
            status = red(_("failed to delete"))
        else:
            status = red(_("failed"))
        if show_diff and not created and not deleted:
            output = "{x} {node}  {bundle}  {item}  {status}\n".format(
                bundle=bold(bundle),
                item=item,
                node=bold(node),
                status=status,
                x=bold(red("✘")),
            )
            diff = diff_dict(details[1], details[0], skip_missing_in_target=True)
            output += prefix_lines(diff, f"{red('│')} ")
            output += red("╵")
            return output
        else:
            return "{x} {node}  {bundle}  {item}  {status} {details}".format(
                bundle=bold(bundle),
                details=details_text,
                item=item,
                node=bold(node),
                status=status,
                x=bold(red("✘")),
            )
    elif result == Item.STATUS_ACTION_SUCCEEDED:
        return "{x} {node}  {bundle}  {item}  {status}".format(
            bundle=bold(bundle),
            item=item,
            node=bold(node),
            status=green(_("succeeded")),
            x=bold(green("✓")),
        )
    elif result == Item.STATUS_SKIPPED:
        return "{x} {node}  {bundle}  {item}  {status} {details}".format(
            bundle=bold(bundle),
            details=details_text,
            item=item,
            node=bold(node),
            x=bold(yellow("»")),
            status=yellow(_("skipped")),
        )
    elif result == Item.STATUS_FIXED:
        if created:
            status = green(_("created"))
        elif deleted:
            status = green(_("deleted"))
        else:
            status = green(_("fixed"))
        if not interactive and show_diff:
            output = "{x} {node}  {bundle}  {item}  {status}\n".format(
                bundle=bold(bundle),
                item=item,
                node=bold(node),
                x=bold(green("✓")),
                status=status,
            )
            diff = "\n"
            if created:
                diff += dict_to_text(details, value_color=green)
            elif deleted:
                diff += dict_to_text(details, value_color=red)
            else:
                diff += diff_dict(details[1], details[0], skip_missing_in_target=True)
            output += prefix_lines(diff, f"{green('│')} ")
            output += green("╵")
            return output
        else:
            return "{x} {node}  {bundle}  {item}  {status}".format(
                bundle=bold(bundle),
                item=item,
                node=bold(node),
                x=bold(green("✓")),
                status=status,
            )


class Node:
    OS_FAMILY_BSD = (
        'freebsd',
        'macos',
        'netbsd',
        'openbsd',
    )
    OS_FAMILY_DEBIAN = (
        'debian',
        'ubuntu',
        'raspbian',
    )
    OS_FAMILY_REDHAT = (
        'rhel',
        'centos',
        'fedora',
        'oraclelinux',
    )

    OS_FAMILY_LINUX = (
        'alpine',
        'amazonlinux',
        'arch',
        'opensuse',
        'openwrt',
        'gentoo',
        'linux',
    ) + \
        OS_FAMILY_DEBIAN + \
        OS_FAMILY_REDHAT

    OS_FAMILY_UNIX = OS_FAMILY_BSD + OS_FAMILY_LINUX
    OS_KNOWN = OS_FAMILY_UNIX + ('kubernetes', 'routeros')

    def __init__(self, name, attributes=None, repo=None):
        if attributes is None:
            attributes = {}

        if not validate_name(name):
            raise RepositoryError(_("'{}' is not a valid node name").format(name))

        with error_context(node_name=name):
            validate_dict(attributes, NODE_ATTR_TYPES)

        attributes = normalize_dict(attributes, GROUP_ATTR_TYPES_ENFORCED)

        self._add_host_keys = environ.get('BW_ADD_HOST_KEYS', False) == "1"
        self._attributes = attributes
        self._dynamic_attribute_cache = {}
        self._ssh_conn_established = False
        self._ssh_first_conn_lock = Lock()
        self.file_path = attributes.get('file_path')
        self.hostname = attributes.get('hostname', name)
        self.name = name
        self.repo = repo

        self.convert_magic_strings()

        for attr in GROUP_ATTR_DEFAULTS:
            setattr(self, "_{}".format(attr), self._attributes.get(attr))

    def __getattr__(self, name):
        with suppress(KeyError):
            return self._dynamic_attribute_cache[name]
        try:
            func = self.repo.node_attribute_functions[name]
        except KeyError:
            raise AttributeError(name)
        else:
            with error_context(node=self.name, attr=name):
                value = func(self)
            self._dynamic_attribute_cache[name] = value
            return value

    def __lt__(self, other):
        return self.name < other.name

    def __repr__(self):
        return "<Node '{}'>".format(self.name)

    def __str__(self):
        return self.name

    @cached_property_set
    def bundles(self):
        with io.job(_("{node}  loading bundles").format(node=bold(self.name))):
            bundle_names = set(self._attributes.get('bundles', set()))

            for group in self.groups:
                for bundle_name in set(group._attributes.get('bundles', set())):
                    bundle_names.add(bundle_name)

            for bundle_name in bundle_names:
                try:
                    yield Bundle(self, bundle_name)
                except NoSuchBundle:
                    raise NoSuchBundle(_(
                        "Node '{node}' wants bundle '{bundle}', but it doesn't exist."
                    ).format(
                        bundle=bundle_name,
                        node=self.name,
                    ))

    @cached_property
    def expected_state(self):
        node_dict = {}
        for item in self.items:
            if item.ITEM_TYPE_NAME != 'action':  # actions have no expected_state
                node_dict[item.id] = item.hash()
        return node_dict

    def covered_by_autoskip_selector(self, autoskip_selector):
        """
        True if this node should be skipped based on the given selector
        string (e.g. "node:foo,group:bar").
        """
        components = [c.strip() for c in autoskip_selector]
        if "node:{}".format(self.name) in components:
            return True
        for group in self.groups:
            if "group:{}".format(group.name) in components:
                return True
        return False

    @io.job_wrapper(_("{}  converting magic strings").format(bold("{0.name}")))
    def convert_magic_strings(self):
        # Lives in its own function so we can use `io.job_wrapper()`
        self._attributes = convert_magic_strings(self.repo, self._attributes)

    def group_membership_hash(self):
        return hash_state_dict(sorted(names(self.groups)))

    @property
    def immediate_groups(self):
        result = set()
        for group_name in self._attributes.get('groups', set()):
            try:
                result.add(self.repo.get_group(group_name))
            except NoSuchGroup:
                raise RepositoryError(_(
                    "Node '{node}' has '{group}' listed as a group, "
                    "but no such group could be found."
                ).format(
                    node=self.name,
                    group=group_name,
                ))
        for group in self.repo.groups:
            if group in result:
                continue
            elif self in group._nodes_from_members:
                result.add(self.repo.get_group(group.name))
            else:
                for pattern in group._member_patterns:
                    if pattern.search(self.name) is not None:
                        result.add(self.repo.get_group(group.name))
        return result

    @cached_property_set
    @io.job_wrapper(_("{}  determining groups").format(bold("{0.name}")))
    def groups(self):
        for group in self.immediate_groups:
            yield group
            for parent_group in group.parent_groups:
                # these are not just the *immediate* parent groups
                yield parent_group

    def has_any_bundle(self, bundle_list):
        for bundle_name in bundle_list:
            if self.has_bundle(bundle_name):
                return True
        return False

    def has_bundle(self, bundle_name):
        for bundle in self.bundles:
            if bundle.name == bundle_name:
                return True
        return False

    def hash(self):
        return hash_state_dict(self.expected_state)

    def in_any_group(self, group_list):
        for group_name in group_list:
            if self.in_group(group_name):
                return True
        return False

    def in_group(self, group_name):
        for group in self.groups:
            if group.name == group_name:
                return True
        return False

    @cached_property_set
    def items(self):
        items = {}
        if not self.dummy:
            for bundle in self.bundles:
                for item in bundle.items:
                    if item.id in items:
                        raise BundleError(_(
                            "duplicate definition of {item} in bundles '{bundle1}' and '{bundle2}'"
                        ).format(
                            item=item.id,
                            bundle1=item.bundle.name,
                            bundle2=items[item.id].bundle.name,
                        ))
                    else:
                        items[item.id] = item
        return items.values()

    @cached_property
    def magic_number(self):
        return int(md5(self.name.encode('UTF-8')).hexdigest(), 16)

    def check_connection(self):
        try:
            # Running "true" is meant to catch connection errors early,
            # but this only works on UNIX-y systems (i.e., not k8s).
            if self.os in self.OS_FAMILY_UNIX:
                self.run("true")
            elif self.os == 'routeros':
                self.run_routeros("/nothing")
        except (RemoteException, TransportException) as exc:
            io.stdout(_("{x} {node}  Connection error: {msg}").format(
                msg=exc,
                node=bold(self.name),
                x=red("!"),
            ))
            return False
        else:
            return True

    def apply(
        self,
        autoskip_selector=(),
        autoonly_selector=(),
        interactive=False,
        force=False,
        show_diff=True,
        show_skipped_items=True,
        skip_list=(),
        workers=4,
        add_lock_for_fixed_items=False,
        lock_expiry="8h",
    ):
        if not list(self.items):
            io.stdout(_("{x} {node}  has no items").format(
                node=bold(self.name),
                x=yellow("»"),
            ))
            return None

        if self.covered_by_autoskip_selector(autoskip_selector):
            io.stdout(_("{x} {node}  skipped by --skip").format(
                node=bold(self.name),
                x=yellow("»"),
            ))
            return None

        if self.name in skip_list:
            io.stdout(_("{x} {node}  skipped by --resume-file").format(
                node=bold(self.name),
                x=yellow("»"),
            ))
            return None

        try:
            self.repo.hooks.node_apply_start(
                repo=self.repo,
                node=self,
                interactive=interactive,
            )
        except SkipNode as exc:
            io.stdout(_("{x} {node}  skipped by hook ({reason})").format(
                node=bold(self.name),
                reason=str(exc) or _("no reason given"),
                x=yellow("»"),
            ))
            return None

        start = datetime.now()
        io.stdout(_("{x} {node}  {started} at {time}").format(
            node=bold(self.name),
            started=bold(_("started")),
            time=start.strftime("%Y-%m-%d %H:%M:%S"),
            x=blue("i"),
        ))

        error = None
        if not self.check_connection():
            error = _("Connection error (details above)")
            item_results = []
        else:
            try:
                with NodeLock(self, interactive=interactive, ignore=force) as lock:
                    item_results = apply_items(
                        self,
                        autoskip_selector=autoskip_selector,
                        autoonly_selector=autoonly_selector,
                        my_soft_locks=lock.my_soft_locks,
                        other_peoples_soft_locks=lock.other_peoples_soft_locks,
                        workers=workers,
                        interactive=interactive,
                        show_diff=show_diff,
                        show_skipped_items=show_skipped_items,
                    )

                    if add_lock_for_fixed_items:
                        items_to_lock = set()

                        for item in item_results:
                            item_id, status_code, _duration = item
                            if status_code in (Item.STATUS_FIXED, Item.STATUS_ACTION_SUCCEEDED):
                                items_to_lock.add(item_id)

                        if items_to_lock:
                            lock_id = "APPLY_" + randstr(length=4).upper()
                            softlock_add(
                                self,
                                lock_id,
                                item_selectors=sorted(items_to_lock),
                                expiry=lock_expiry,
                            )
            except NodeLockedException as e:
                if not interactive:
                    io.stderr(_(
                        "{x} {node} already locked by {user} at {date} ({duration} ago, "
                        "`bw apply -f` to override)"
                    ).format(
                        date=bold(e.args[0]['date']),
                        duration=e.args[0]['duration'],
                        node=bold(self.name),
                        user=bold(e.args[0]['user']),
                        x=red("!"),
                    ))
                error = _("Node locked (details above)")
                item_results = []
        result = ApplyResult(self, item_results)
        result.start = start
        result.end = datetime.now()

        io.stdout(_("{x} {node}  {completed} after {time}  ({stats})").format(
            completed=bold(_("completed")),
            node=bold(self.name),
            stats=format_node_result(result),
            time=format_duration(result.end - start),
            x=blue("i"),
        ))

        self.repo.hooks.node_apply_end(
            repo=self.repo,
            node=self,
            duration=result.duration,
            interactive=interactive,
            result=result,
        )

        if error:
            raise GracefulApplyException(error)
        else:
            return result

    def download(self, remote_path, local_path):
        return operations.download(
            self.hostname,
            remote_path,
            local_path,
            add_host_keys=self._add_host_keys,
            username=self.username,
            wrapper_inner=self.cmd_wrapper_inner,
            wrapper_outer=self.cmd_wrapper_outer,
        )

    def get_item(self, item_id):
        return find_item(item_id, self.items)

    @property
    def metadata(self):
        return self.repo._metadata_proxy_for_node(self.name)

    def metadata_hash(self):
        return hash_metadata(self.metadata)

    @property
    def metadata_defaults(self):
        for bundle in self.bundles:
            if bundle._metadata_defaults_and_reactors[0]:
                yield (
                    "metadata_defaults:{}".format(bundle.name),
                    bundle._metadata_defaults_and_reactors[0],
                )

    @property
    def metadata_reactors(self):
        for bundle in self.bundles:
            for reactor in bundle._metadata_defaults_and_reactors[1]:
                yield (
                    "metadata_reactor:{}.{}".format(
                        bundle.name,
                        reactor.__name__,
                    ),
                    reactor,
                )

    def rename(self, new_name):
        if not self.is_toml:
            raise ValueError(_(
                "{node} cannot be renamed: not in TOML format"
            ).format(node=self.name))
        new_path = join(dirname(self.file_path), f"{new_name}.toml")
        rename(self.file_path, new_path)
        self.file_path = new_path
        self.name = new_name

    def run(self, command, data_stdin=None, may_fail=False, log_output=False, user="root"):
        assert self.os in self.OS_FAMILY_UNIX

        if log_output:
            def log_function(msg):
                io.stdout("{x} {node}  {msg}".format(
                    node=bold(self.name),
                    msg=force_text(msg).rstrip("\n"),
                    x=cyan("›"),
                ))
        else:
            log_function = None

        if not self._ssh_conn_established:
            # Sometimes we're opening SSH connections to a node too fast
            # for OpenSSH to establish the ControlMaster socket for the
            # second and following connections to use.
            # To prevent this, we just wait until a first dummy command
            # has completed on the node before trying to reuse the
            # multiplexed connection.
            if self._ssh_first_conn_lock.acquire(False):
                self.repo.hooks.node_ssh_connect(
                    repo=self.repo,
                    node=self,
                )
                try:
                    with io.job(_("{}  establishing connection...").format(bold(self.name))):
                        operations.run(
                            self.hostname,
                            "true",
                            add_host_keys=self._add_host_keys,
                            username=self.username,
                        )
                    self._ssh_conn_established = True
                finally:
                    self._ssh_first_conn_lock.release()
            else:
                # we didn't get the lock immediately, now we just wait
                # until it is released before we proceed
                with self._ssh_first_conn_lock:
                    pass

        return operations.run(
            self.hostname,
            command,
            add_host_keys=self._add_host_keys,
            data_stdin=data_stdin,
            ignore_failure=may_fail,
            log_function=log_function,
            username=self.username,
            wrapper_inner=self.cmd_wrapper_inner,
            wrapper_outer=self.cmd_wrapper_outer,
            user=user,
        )

    def run_ipmitool(self, command, log_output=False):
        if not (self.ipmi_hostname and self.ipmi_username and self.ipmi_password):
            raise ValueError(_("node {} has no or invalid ipmi configuration").format(self.name))

        if log_output:
            def log_function(msg):
                io.stdout("{x} {node}  {msg}".format(
                    node=bold(self.name),
                    msg=force_text(msg).rstrip("\n"),
                    x=cyan("›"),
                ))
        else:
            log_function = None

        return operations.run_ipmitool(
            # these might be Faults
            self.ipmi_hostname,
            str(self.ipmi_username),
            str(self.ipmi_password),
            command,
            interface=self.ipmi_interface,
            log_function=log_function,
        )

    def run_routeros(self, *command):
        assert self.os == 'routeros'
        return operations.run_routeros(
            self.hostname,
            self.username,
            self.password,
            *command,
        )

    @property
    def is_toml(self):
        return self.file_path and self.file_path.endswith(".toml")

    @cached_property
    def toml(self):
        if not self.is_toml:
            raise ValueError(_("node {} not in TOML format").format(self.name))
        return toml_parse(get_file_contents(self.file_path))

    def toml_save(self):
        try:
            toml_doc = self.toml
        except ValueError:
            attributes = self._attributes.copy()
            del attributes['file_path']
            toml_doc = dict_to_toml(attributes)
            self.file_path = join(self.repo.path, "nodes", self.name + ".toml")
        if not exists(join(self.repo.path, "nodes")):
            mkdir(join(self.repo.path, "nodes"))
        with open(self.file_path, 'w') as f:
            f.write(toml_clean(toml_dump(toml_doc)))

    def toml_set(self, path, value):
        if not isinstance(path, tuple):
            path = path.split("/")
        set_key_at_path(self.toml, path, value)

    def upload(self, local_path, remote_path, mode=None, owner="", group="", may_fail=False):
        assert self.os in self.OS_FAMILY_UNIX
        return operations.upload(
            self.hostname,
            local_path,
            remote_path,
            add_host_keys=self._add_host_keys,
            group=group,
            mode=mode,
            owner=owner,
            ignore_failure=may_fail,
            username=self.username,
            wrapper_inner=self.cmd_wrapper_inner,
            wrapper_outer=self.cmd_wrapper_outer,
        )

    def verify(
        self,
        autoskip_selector=(),
        autoonly_selector=(),
        show_all=False,
        show_diff=True,
        workers=4,
    ):
        result = []
        start = datetime.now()

        if not self.items:
            io.stdout(_("{x} {node}  has no items").format(node=bold(self.name), x=yellow("!")))
        elif not self.check_connection():
            io.stdout(_("{x} {node}  Connection error (details above)").format(node=bold(self.name), x=red("!")))
        else:
            result = verify_items(
                self,
                autoskip_selector=autoskip_selector,
                autoonly_selector=autoonly_selector,
                show_all=show_all,
                show_diff=show_diff,
                workers=workers,
            )

        return {
            'good': result.count(True),
            'bad': result.count(False),
            'unknown': result.count(None),
            'duration': datetime.now() - start,
        }


def build_attr_property(attr, default):
    def method(self):
        attr_source = None
        attr_value = None
        group_order = [
            self.repo.get_group(group_name)
            for group_name in _flatten_group_hierarchy(self.groups)
        ]

        for group in group_order:
            if getattr(group, attr) is not None:
                attr_source = "group:{}".format(group.name)
                attr_value = getattr(group, attr)

        if getattr(self, "_{}".format(attr)) is not None:
            attr_source = "node"
            attr_value = getattr(self, "_{}".format(attr))

        if attr_value is None:
            attr_source = "default"
            attr_value = default

        io.debug(_("node {node} gets its {attr} attribute from: {source}").format(
            node=self.name,
            attr=attr,
            source=attr_source,
        ))
        return attr_value
    method.__name__ = "_group_attr_{}".format(attr)  # required for cached_property
    return cached_property(method)

for attr, default in GROUP_ATTR_DEFAULTS.items():
    setattr(Node, attr, build_attr_property(attr, default))


def verify_items(
    node,
    autoskip_selector=(),
    autoonly_selector=(),
    show_all=False,
    show_diff=True,
    workers=1,
):
    items = []
    for item in node.items:
        if not item.triggered:
            items.append(item)
        else:
            io.progress_advance()

    try:
        # See comment in node.apply().
        if node.os in node.OS_FAMILY_UNIX:
            node.run("true")
    except (RemoteException, TransportException) as exc:
        io.stdout(_("{x} {node}  Connection error: {msg}").format(
            msg=exc,
            node=bold(node.name),
            x=red("!"),
        ))
        for item in items:
            io.progress_advance()
        return [None for item in items]

    def tasks_available():
        return bool(items)

    def next_task():
        while True:
            try:
                item = items.pop()
            except IndexError:
                return None
            if item._faults_missing_for_attributes:
                if item.error_on_missing_fault:
                    item._raise_for_faults()
                else:
                    io.progress_advance()
                    io.stdout(_("{x} {node}  {bundle}  {item}  ({msg})").format(
                        bundle=bold(item.bundle.name),
                        item=item.id,
                        msg=yellow(_("Fault unavailable")),
                        node=bold(node.name),
                        x=yellow("»"),
                    ))
            else:
                return {
                    'task_id': node.name + ":" + item.bundle.name + ":" + item.id,
                    'target': item.verify,
                    'kwargs': {
                        'autoskip_selector': autoskip_selector,
                        'autoonly_selector': autoonly_selector,
                    },
                }

    def handle_exception(task_id, exception, traceback):
        node_name, bundle_name, item_id = task_id.split(":", 2)
        io.progress_advance()
        if isinstance(exception, (ItemSkipped, NotImplementedError)):
            pass
        else:
            # Unlike with `bw apply`, it is OK for `bw verify` to encounter
            # exceptions when getting an item's status. `bw verify` doesn't
            # care about dependencies and therefore cannot know that looking
            # up a database user requires the database to be installed in
            # the first place.
            item = node.get_item(task_id.split(":", 2)[2])
            output = _("{x} {node}  {bundle}  {item}  (unable to get status)\n").format(
                bundle=bold(bundle_name),
                item=item_id,
                node=bold(node_name),
                x=bold(cyan("?")),
            )
            output += prefix_lines(traceback, cyan("│ "))
            output += prefix_lines(repr(exception), cyan("│ "))
            if item._command_results:
                output += format_item_command_results(item._command_results)
                # free up memory
                del item._command_results
            io.stderr(output)
        return None  # count this result as "unknown"

    def handle_result(task_id, return_value, duration):
        io.progress_advance()
        unless_result, item_status, display = return_value
        node_name, bundle_name, item_id = task_id.split(":", 2)
        if not unless_result and not item_status.correct:
            if item_status.must_be_created:
                details_text = red(_("missing"))
            elif item_status.must_be_deleted:
                details_text = red(_("found"))
            elif show_diff:
                details_text = ""
            else:
                details_text = ", ".join(sorted(display[2]))
            if show_diff:
                diff = "\n"
                if item_status.must_be_created:
                    diff += dict_to_text(display, value_color=green)
                elif item_status.must_be_deleted:
                    diff += dict_to_text(display, value_color=red)
                else:
                    diff += diff_dict(display[1], display[0], skip_missing_in_target=True)
                output = "{x} {node}  {bundle}  {item}  {details}\n".format(
                    bundle=bold(bundle_name),
                    details=details_text,
                    item=item_id,
                    node=bold(node_name),
                    x=red("✘"),
                )
                output += prefix_lines(diff, red('│ '))
                io.stderr(output + red("╵"))
            else:
                io.stderr("{x} {node}  {bundle}  {item}  {details}".format(
                    bundle=bold(bundle_name),
                    details=details_text,
                    item=item_id,
                    node=bold(node_name),
                    x=red("✘"),
                ))
            return False
        else:
            if show_all:
                io.stdout("{x} {node}  {bundle}  {item}".format(
                    bundle=bold(bundle_name),
                    item=item_id,
                    node=bold(node_name),
                    x=green("✓"),
                ))
            return True

    worker_pool = WorkerPool(
        tasks_available,
        next_task,
        handle_result,
        handle_exception=handle_exception,
        pool_id="verify_{}".format(node.name),
        workers=workers,
    )
    return worker_pool.run()



from atexit import register as at_exit
from base64 import b64decode
from collections import defaultdict
from contextlib import contextmanager, suppress
from datetime import datetime

try:
    from functools import cache
except ImportError:  # Python 3.8
    cache = lambda f: f
from hashlib import md5
from os import getenv, getpid, makedirs, mkdir, rmdir
from os.path import basename, dirname, exists, isfile, join, normpath
from shlex import quote
from shutil import rmtree
from subprocess import check_output, CalledProcessError, STDOUT
from sys import exc_info
from tempfile import gettempdir
from time import sleep
from traceback import format_exception

from jinja2 import Environment, FileSystemLoader
from mako.lookup import TemplateLookup
from mako.template import Template
from requests import head

from bundlewrap.exceptions import BundleError, FaultUnavailable, TemplateError
from bundlewrap.items import BUILTIN_ITEM_ATTRIBUTES, Item
from bundlewrap.items.directories import validator_mode
from bundlewrap.utils import cached_property, download, hash_local_file, sha256, tempfile
from bundlewrap.utils.remote import PathInfo
from bundlewrap.utils.text import bold, force_text, mark_for_translation as _
from bundlewrap.utils.text import is_subdirectory
from bundlewrap.utils.ui import io

DIFF_MAX_FILE_SIZE = 1024 * 1024 * 5  # bytes


@cache
def check_download(url, timeout):
    try:
        head(url, allow_redirects=True, timeout=timeout).raise_for_status()
    except Exception as exc:
        return exc
    else:
        return None


def content_processor_base64(item):
    # .encode() is required for pypy3 only
    return b64decode(item._template_content.encode())


def content_processor_jinja2(item):
    loader = FileSystemLoader(searchpath=[item.item_data_dir, item.item_dir])
    env = Environment(loader=loader)

    template = env.from_string(item._template_content)

    io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: rendering with Jinja2...")
    start = datetime.now()
    try:
        content = template.render(
            item=item,
            bundle=item.bundle,
            node=item.node,
            repo=item.node.repo,
            **item.attributes['context']
        )
    except FaultUnavailable as e:
        io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: Fault Unavailable - {e}")
        raise
    except Exception as e:
        io.stderr("".join(format_exception(*exc_info())))
        raise TemplateError(_(
            "Error while rendering template for {node}:{bundle}:{item}: {error}"
        ).format(
            bundle=item.bundle.name,
            error=e,
            item=item.id,
            node=item.node.name,
        ))
    duration = datetime.now() - start
    io.debug("{node}:{bundle}:{item}: rendered in {time:.09f} s".format(
        bundle=item.bundle.name,
        item=item.id,
        node=item.node.name,
        time=duration.total_seconds(),
    ))
    return content.encode(item.attributes['encoding'])


def content_processor_mako(item):
    template = Template(
        item._template_content.encode('utf-8'),
        input_encoding='utf-8',
        lookup=TemplateLookup(directories=[item.item_data_dir, item.item_dir]),
        output_encoding=item.attributes['encoding'],
    )
    io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: rendering with Mako...")
    start = datetime.now()
    try:
        content = template.render(
            item=item,
            bundle=item.bundle,
            node=item.node,
            repo=item.node.repo,
            **item.attributes['context']
        )
    except FaultUnavailable as e:
        io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: Fault Unavailable - {e}")
        raise
    except Exception as e:
        io.stderr("".join(format_exception(*exc_info())))
        if isinstance(e, NameError) and str(e) == "Undefined":
            # Mako isn't very verbose here. Try to give a more useful
            # error message - even though we can't pinpoint the excat
            # location of the error. :/
            e = _("Undefined variable (look for '${...}')")
        elif isinstance(e, KeyError):
            e = _("KeyError: {}").format(str(e))
        raise TemplateError(_(
            "Error while rendering template for {node}:{bundle}:{item}: {error}"
        ).format(
            bundle=item.bundle.name,
            error=e,
            item=item.id,
            node=item.node.name,
        ))
    duration = datetime.now() - start
    io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: rendered in {duration.total_seconds():.09f} s")
    return content


def content_processor_text(item):
    return item._template_content.encode(item.attributes['encoding'])


CONTENT_PROCESSORS = {
    'any': lambda item: b"",
    'base64': content_processor_base64,
    'binary': None,
    'jinja2': content_processor_jinja2,
    'mako': content_processor_mako,
    'text': content_processor_text,
    'download': None,
}


def download_file(item):
    file_name_hashed = md5(item.attributes['source'].encode('UTF-8')).hexdigest()

    cache_path = getenv("BW_FILE_DOWNLOAD_CACHE")
    if cache_path:
        remove_dir = None
        file_path = join(cache_path, file_name_hashed)
        lock_dir = join(cache_path, "{}.bw_lock".format(file_name_hashed))

        makedirs(cache_path, exist_ok=True)
    else:
        remove_dir = join(gettempdir(), "bw-file-download-cache-{}".format(getpid()))
        file_path = join(remove_dir, file_name_hashed)
        lock_dir = join(remove_dir, "{}.bw_lock".format(file_name_hashed))

        makedirs(remove_dir, exist_ok=True)

    io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: download lock dir is {lock_dir}")

    # Since we only download the file once per process, there's no point
    # in displaying the node name here. The file may be used on multiple
    # nodes.
    with io.job(_("{}  {}  waiting for download").format(bold(item.node.name), bold(item.id))):
        while True:
            try:
                mkdir(lock_dir)
                io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: have download lock")
                break
            except FileExistsError:
                io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: waiting for download lock")
                sleep(1)

    try:
        if not isfile(file_path):
            io.debug(
                f"{item.node.name}:{item.bundle.name}:{item.id}: "
                f"starting download from {item.attributes['source']}"
            )
            with io.job(_("{node}  {item}  downloading from {url}").format(
                node=bold(item.node.name),
                item=bold(item.id),
                url=item.attributes['source'],
            )):
                download(
                    item.attributes['source'],
                    file_path,
                    timeout=item.attributes['download_timeout'],
                )
            io.debug(
                f"{item.node.name}:{item.bundle.name}:{item.id}: "
                f"finished download from {item.attributes['source']}"
            )

        # Always do hash verification, if requested.
        if item.attributes['content_hash']:
            with io.job(_("{}  {}  checking file integrity".format(
                bold(item.node.name),
                bold(item.id),
            ))):
                local_hash = hash_local_file(file_path)
                io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: content hash is {local_hash}")
                if local_hash != item.attributes['content_hash']:
                    raise BundleError(_(
                        "could not download correct file from {} - sha256sum mismatch "
                        "(expected {}, got {})"
                    ).format(
                        item.attributes['source'],
                        item.attributes['content_hash'],
                        local_hash
                    ))
                io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: content hash matches")
    finally:
        rmdir(lock_dir)
        io.debug(f"{item.node.name}:{item.bundle.name}:{item.id}: released download lock")

    return file_path, remove_dir


def get_remote_file_contents(node, path):
    """
    Returns the contents of the given path as a string.
    """
    with tempfile() as tmp_file:
        node.download(path, tmp_file)
        with open(tmp_file, 'rb') as f:
            content = f.read()
        return content


def validator_content_type(item_id, value):
    if value not in CONTENT_PROCESSORS:
        raise BundleError(_(
            "invalid content_type for {item}: '{value}'"
        ).format(item=item_id, value=value))


ATTRIBUTE_VALIDATORS = defaultdict(lambda: lambda id, value: None)
ATTRIBUTE_VALIDATORS.update({
    'content_type': validator_content_type,
    'mode': validator_mode,
})


class File(Item):
    """
    A file.
    """
    BUNDLE_ATTRIBUTE_NAME = "files"
    ITEM_ATTRIBUTES = {
        'content': None,
        'content_type': 'text',
        'content_hash': None,
        'context': None,
        'delete': False,
        'download_timeout': 60.0,
        'encoding': "utf-8",
        'group': "root",
        'mode': "0644",
        'owner': "root",
        'source': None,
        'verify_with': None,
        'test_with': None,
    }
    ITEM_TYPE_NAME = "file"

    def __repr__(self):
        return "<File path:{} content_type:{} owner:{} group:{} mode:{} delete:{}>".format(
            quote(self.name),
            self.attributes['content_type'],
            self.attributes['owner'],
            self.attributes['group'],
            self.attributes['mode'],
            self.attributes['delete'],
        )

    @property
    def _template_content(self):
        if self.attributes['source'] is not None:
            filename = join(self.item_data_dir, self.attributes['source'])
            if not exists(filename):
                filename = join(self.item_dir, self.attributes['source'])
            with open(filename, 'rb') as f:
                return force_text(f.read())
        else:
            return force_text(self.attributes['content'])

    @cached_property
    def content(self):
        return CONTENT_PROCESSORS[self.attributes['content_type']](self)

    @cached_property
    def content_hash(self):
        if self.attributes['content_type'] in ('binary', 'download'):
            return hash_local_file(self.template)
        else:
            return sha256(self.content)

    @cached_property
    def template(self):
        if self.attributes['content_type'] == 'download':
            file_path, remove_dir = download_file(self)
            if remove_dir:
                io.debug(_("registering {} for deletion on exit").format(remove_dir))
                at_exit(rmtree, remove_dir, ignore_errors=True)
            return file_path
        data_template = join(self.item_data_dir, self.attributes['source'])
        if exists(data_template):
            return data_template
        return join(self.item_dir, self.attributes['source'])

    @property
    def expected_state(self):
        if self.attributes['delete']:
            return None

        state = {'type': 'file'}
        if self.attributes['content_type'] != 'any':
            if self.attributes['content_type'] == 'download' and self.attributes['content_hash']:
                state['content_hash'] = self.attributes['content_hash']
            else:
                state['content_hash'] = self.content_hash

        for optional_attr in ('group', 'mode', 'owner'):
            if self.attributes[optional_attr] is not None:
                state[optional_attr] = self.attributes[optional_attr]

        return state

    def fix(self, status):
        if status.must_be_created or status.must_be_deleted or 'type' in status.keys_to_fix:
            self._fix_type(status)
        else:
            for fix_type in ('content_hash', 'mode', 'owner', 'group'):
                if fix_type in status.keys_to_fix:
                    if fix_type == 'group' and \
                        'owner' in status.keys_to_fix:
                        # owner and group are fixed with a single chown
                        continue
                    if fix_type in ('mode', 'owner', 'group') and \
                        'content' in status.keys_to_fix:
                        # fixing content implies settings mode and owner/group
                        continue
                    getattr(self, "_fix_" + fix_type)(status)

    def _fix_content_hash(self, status):
        with self._write_local_file() as local_path:
            with io.job(_("{}  {}  uploading to node").format(
                bold(self.node.name),
                bold(self.id),
            )):
                self.node.upload(
                    local_path,
                    self.name,
                    mode=self.attributes['mode'],
                    owner=self.attributes['owner'] or "",
                    group=self.attributes['group'] or "",
                    may_fail=True,
                )

    def _fix_mode(self, status):
        if self.node.os in self.node.OS_FAMILY_BSD:
            command = "chmod {} {}"
            mode = self.attributes['mode']
        else:
            command = "chmod {} -- {}"
            # GNU chmod refuses to set some modes (e.g., "chmod 0755 dir"
            # when "dir" is a directory which currently has mode "2755")
            # unless there's an additional leading zero.
            mode = '0' + self.attributes['mode'].zfill(4)
        self.run(command.format(mode, quote(self.name)))

    def _fix_owner(self, status):
        group = self.attributes['group'] or ""
        if group:
            group = ":" + quote(group)
        if self.node.os in self.node.OS_FAMILY_BSD:
            command = "chown {}{} {}"
        else:
            command = "chown {}{} -- {}"
        self.run(command.format(
            quote(self.attributes['owner'] or ""),
            group,
            quote(self.name),
        ))

    _fix_group = _fix_owner

    def _fix_type(self, status):
        if status.actual_state:
            self.run("rm -rf -- {}".format(quote(self.name)))
        if not status.must_be_deleted:
            self.run("mkdir -p -- {}".format(quote(dirname(self.name))))
            self._fix_content_hash(status)

    def get_auto_attrs(self, items):
        deps = set()
        for item in items:
            if item.ITEM_TYPE_NAME == 'file' and is_subdirectory(item.name, self.name):
                raise BundleError(_(
                    "{item1} (from bundle '{bundle1}') blocking path to "
                    "{item2} (from bundle '{bundle2}')"
                ).format(
                    item1=item.id,
                    bundle1=item.bundle.name,
                    item2=self.id,
                    bundle2=self.bundle.name,
                ))
            elif item.ITEM_TYPE_NAME == 'user' and item.name == self.attributes['owner']:
                if item.attributes['delete']:
                    raise BundleError(_(
                        "{item1} (from bundle '{bundle1}') depends on item "
                        "{item2} (from bundle '{bundle2}') which is set to be deleted"
                    ).format(
                        item1=self.id,
                        bundle1=self.bundle.name,
                        item2=item.id,
                        bundle2=item.bundle.name,
                    ))
                else:
                    deps.add(item.id)
            elif item.ITEM_TYPE_NAME == 'group' and item.name == self.attributes['group']:
                if item.attributes['delete']:
                    raise BundleError(_(
                        "{item1} (from bundle '{bundle1}') depends on item "
                        "{item2} (from bundle '{bundle2}') which is set to be deleted"
                    ).format(
                        item1=self.id,
                        bundle1=self.bundle.name,
                        item2=item.id,
                        bundle2=item.bundle.name,
                    ))
                else:
                    deps.add(item.id)
            elif item.ITEM_TYPE_NAME == "zfs_dataset":
                if item.attributes.get('mountpoint') not in (None, "none"):
                    if is_subdirectory(item.attributes['mountpoint'], self.name):
                        deps.add(item.id)
            elif item.ITEM_TYPE_NAME in ('directory', 'symlink'):
                if is_subdirectory(item.name, self.name):
                    deps.add(item.id)
        return {
            'needs': deps,
        }

    @property
    def actual_state(self):
        use_uid = self.attributes['owner'] is not None and self.attributes['owner'].startswith('+')
        use_gid = self.attributes['group'] is not None and self.attributes['group'].startswith('+')
        path_info = PathInfo(self.node, self.name)
        if not path_info.exists:
            return None
        else:
            return {
                'type': 'file' if path_info.is_file else path_info.stat['type'],
                'content_hash': path_info.sha256 if path_info.is_file else None,
                'mode': path_info.mode,
                'owner': ('+' + path_info.owner_id) if use_uid else path_info.owner,
                'group': ('+' + path_info.group_id) if use_gid else path_info.group,
                'size': path_info.size,
            }

    def display_on_create(self, expected_state):
        if (
            self.attributes['content_type'] not in ('any', 'base64', 'binary', 'download') and
            len(self.content) < DIFF_MAX_FILE_SIZE
        ):
            del expected_state['content_hash']
            expected_state['content'] = force_text(self.content)
        if self.attributes['content_type'] == 'download':
            expected_state['source'] = self.attributes['source']
        del expected_state['type']
        return expected_state

    def display_on_fix(self, expected_state, actual_state, keys):
        if (
            'content_hash' in keys and
            self.attributes['content_type'] not in ('base64', 'binary', 'download') and
            actual_state['size'] < DIFF_MAX_FILE_SIZE and
            len(self.content) < DIFF_MAX_FILE_SIZE and
            PathInfo(self.node, self.name).is_text_file
        ):
            keys.remove('content_hash')
            keys.add('content')
            del expected_state['content_hash']
            del actual_state['content_hash']
            expected_state['content'] = self.content.decode(self.attributes['encoding'])
            actual_state['content'] = get_remote_file_contents(
                self.node,
                self.name,
            ).decode('utf-8', 'backslashreplace')
        if 'type' in keys:
            keys.discard('content_hash')
        if self.attributes['content_type'] == 'download':
            expected_state['source'] = self.attributes['source']
            actual_state['source'] = ''
        if actual_state:
            del actual_state['size']
            if self.attributes['content_type'] == 'any':
                with suppress(KeyError):
                    del actual_state['content_hash']
        return (expected_state, actual_state, keys)

    def display_on_delete(self, actual_state):
        del actual_state['content_hash']
        path_info = PathInfo(self.node, self.name)
        if (
            actual_state['size'] < DIFF_MAX_FILE_SIZE and
            path_info.is_text_file
        ):
            actual_state['content'] = get_remote_file_contents(self.node, self.name)
        if path_info.is_file:
            actual_state['size'] = f"{actual_state['size']} bytes"
        return actual_state

    def patch_attributes(self, attributes):
        if (
            'content' not in attributes and
            'source' not in attributes and
            attributes.get('content_type', 'text') != 'any' and
            attributes.get('delete', False) is False
        ):
            attributes['source'] = basename(self.name)
        if 'context' not in attributes:
            attributes['context'] = {}
        if 'mode' in attributes and attributes['mode'] is not None:
            attributes['mode'] = str(attributes['mode']).zfill(4)
        if 'group' not in attributes and self.node.os in self.node.OS_FAMILY_BSD:
            # BSD doesn't have a root group, so we have to use a
            # different default value here
            attributes['group'] = 'wheel'
        return attributes

    def preview(self):
        if (
            self.attributes['content_type'] in ('any', 'base64', 'binary') or
            self.attributes['delete'] is True
        ):
            raise ValueError
        return self.content.decode(self.attributes['encoding'])

    def test(self):
        if (
            self.attributes['source']
            and self.attributes['content_type'] != 'download'
            and not exists(self.template)
        ):
            raise BundleError(_(
                "{item} from bundle '{bundle}' refers to missing "
                "file '{path}' in its 'source' attribute"
            ).format(
                bundle=self.bundle.name,
                item=self.id,
                path=self.template,
            ))

        if (
            self.attributes['delete']
            or self.attributes['content_type'] == 'any'
        ):
            pass
        elif (
            self.attributes['content_type'] == 'download'
            and not self.attributes['content_hash']
        ):
            download_exc = check_download(
                self.attributes['source'],
                self.attributes['download_timeout'],
            )
            if download_exc is not None:
                raise download_exc
        else:
            with self._write_local_file() as local_path:
                if self.attributes['test_with']:
                    cmd = self.attributes['test_with'].format(quote(local_path))
                    exitcode, stdout = self._run_validator(cmd)
                    if exitcode == 0:
                        io.debug(f"{self.id} passed local validation")
                    else:
                        raise BundleError(_(
                            "{i} failed local validation using: {c}\n\n{out}"
                        ).format(
                            c=cmd,
                            i=self.id,
                            out=stdout,
                        ))

    def _run_validator(self, cmd):
        io.debug(f"calling local validator for {self.node.name}:{self.id}: {cmd}")
        try:
            p = check_output(cmd, shell=True, stderr=STDOUT)
        except CalledProcessError as e:
            return e.returncode, e.output.decode()
        else:
            return 0, p.decode()

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        if attributes.get('delete', False):
            for attr in attributes.keys():
                if attr not in ['delete'] + list(BUILTIN_ITEM_ATTRIBUTES.keys()):
                    raise BundleError(_(
                        "{item} from bundle '{bundle}' cannot have other "
                        "attributes besides 'delete'"
                    ).format(item=item_id, bundle=bundle.name))

        if 'content' in attributes and 'source' in attributes:
            raise BundleError(_(
                "{item} from bundle '{bundle}' cannot have both 'content' and 'source'"
            ).format(item=item_id, bundle=bundle.name))

        if 'content' in attributes and attributes.get('content_type') == 'binary':
            raise BundleError(_(
                "{item} from bundle '{bundle}' cannot have binary inline content "
                "(use content_type 'base64' instead)"
            ).format(item=item_id, bundle=bundle.name))

        if 'content_hash' in attributes and attributes.get('content_type') != 'download':
            raise BundleError(_(
                "{item} from bundle '{bundle}' specified 'content_hash', but is "
                "not of type 'download'"
            ).format(item=item_id, bundle=bundle.name))

        if 'download_timeout' in attributes and attributes.get('content_type') != 'download':
            raise BundleError(_(
                "{item} from bundle '{bundle}' specified 'download_timeout', but is "
                "not of type 'download'"
            ).format(item=item_id, bundle=bundle.name))

        if 'download_timeout' in attributes:
            if (
                not isinstance(attributes['download_timeout'], float)
                or attributes['download_timeout'] <= 0.0
            ):
                raise BundleError(_(
                    "download_timeout for {item} from bundle '{bundle}' must be a float > 0.0"
                ).format(item=item_id, bundle=bundle.name))

        if attributes.get('content_type') == 'download':
            if 'source' not in attributes:
                raise BundleError(_(
                    "{item} from bundle '{bundle}' is of type 'download', but missing "
                    "required attribute 'source'"
                ).format(item=item_id, bundle=bundle.name))
            elif '://' not in attributes['source']:
                raise BundleError(_(
                    "{item} from bundle '{bundle}' is of type 'download', but {source} "
                    "does not look like a URL"
                ).format(item=item_id, bundle=bundle.name, source=attributes['source']))

        if 'encoding' in attributes and attributes.get('content_type') in (
                'any',
                'base64',
                'binary',
                'download',
        ):
            raise BundleError(_(
                "content_type of {item} from bundle '{bundle}' cannot provide different encoding "
                "(remove the 'encoding' attribute)"
            ).format(item=item_id, bundle=bundle.name))

        if (
            attributes.get('content_type', None) == "any" and (
            'content' in attributes or
            'encoding' in attributes or
            'source' in attributes
        )
        ):
            raise BundleError(_(
                "{item} from bundle '{bundle}' with content_type 'any' "
                "must not define 'content', 'encoding' and/or 'source'"
            ).format(item=item_id, bundle=bundle.name))

        for key, value in attributes.items():
            ATTRIBUTE_VALIDATORS[key](item_id, value)

    @classmethod
    def validate_name(cls, bundle, name):
        if normpath(name) == "/":
            raise BundleError(_("'/' cannot be a file"))
        if normpath(name) != name:
            raise BundleError(_(
                "'{path}' is an invalid file path, should be '{normpath}' (bundle '{bundle}')"
            ).format(
                bundle=bundle.name,
                normpath=normpath(name),
                path=name,
            ))

    @contextmanager
    def _write_local_file(self):
        """
        Makes the file contents available at the returned temporary path
        and performs local verification if necessary or requested.

        The calling method is responsible for cleaning up the file at
        the returned path (only if not a binary).
        """
        with tempfile() as tmp_file:
            if self.attributes['content_type'] in ('binary', 'download'):
                local_path = self.template
            else:
                local_path = tmp_file
                with open(local_path, 'wb') as f:
                    f.write(self.content)

            if self.attributes['verify_with']:
                cmd = self.attributes['verify_with'].format(quote(local_path))
                exitcode, stdout = self._run_validator(cmd)
                if exitcode == 0:
                    io.debug(f"{self.id} passed local validation")
                else:
                    raise BundleError(_(
                        "{i} failed local validation using: {c}\n\n{out}"
                    ).format(
                        c=cmd,
                        i=self.id,
                        out=stdout,
                    ))

            yield local_path



from contextlib import contextmanager, suppress
from datetime import datetime, timedelta
import faulthandler
from functools import wraps
from os import _exit, environ, getpid, kill
from os.path import join
from select import select
from shutil import get_terminal_size
from signal import signal, SIG_DFL, SIGINT, SIGQUIT, SIGTERM
from subprocess import PIPE, Popen
import sys
import termios
from time import time
from threading import Event, Lock, Thread

from . import STDERR_WRITER, STDOUT_WRITER
from .table import render_table, ROW_SEPARATOR
from .text import (
    HIDE_CURSOR,
    SHOW_CURSOR,
    ansi_clean,
    blue,
    bold,
    format_duration,
    mark_for_translation as _,
    trim_visible_len_to,
)


INFO_EVENT = Event()
QUIT_EVENT = Event()
SHUTDOWN_EVENT_HARD = Event()
SHUTDOWN_EVENT_SOFT = Event()
TTY = STDOUT_WRITER.isatty()


def add_debug_indicator(f):
    @wraps(f)
    def wrapped(self, msg, **kwargs):
        return f(self, "[DEBUG] " + msg, **kwargs)
    return wrapped


def add_debug_timestamp(f):
    @wraps(f)
    def wrapped(self, msg, **kwargs):
        if self.debug_mode:
            msg = f"[{datetime.now().isoformat()}] {msg}"
        return f(self, msg, **kwargs)
    return wrapped


def capture_for_debug_logfile(f):
    @wraps(f)
    def wrapped(self, msg, **kwargs):
        if self.debug_log_file and self._active:
            with self.lock:
                clean_msg = ansi_clean(msg).rstrip("\n")
                self.debug_log_file.write(
                    f"[{datetime.now().isoformat()}] {clean_msg}\n"
                )
        return f(self, msg, **kwargs)
    return wrapped


def clear_formatting(f):
    """
    Makes sure formatting from cut-off lines can't bleed into next one
    """
    @wraps(f)
    def wrapped(self, msg, **kwargs):
        if TTY and environ.get("BW_COLORS", "1") != "0":
            msg = "\033[0m" + msg
        return f(self, msg, **kwargs)
    return wrapped


def sigint_handler(*args, **kwargs):
    """
    This handler is kept short since it interrupts execution of the
    main thread. It's safer to handle these events in their own thread
    because the main thread might be holding the IO lock while it is
    interrupted.
    """
    if not SHUTDOWN_EVENT_SOFT.is_set():
        SHUTDOWN_EVENT_SOFT.set()
    else:
        SHUTDOWN_EVENT_HARD.set()
        faulthandler.dump_traceback()


def sigquit_handler(*args, **kwargs):
    """
    This handler is kept short since it interrupts execution of the
    main thread. It's safer to handle these events in their own thread
    because the main thread might be holding the IO lock while it is
    interrupted.
    """
    INFO_EVENT.set()


def spinner():
    while True:
        for c in "⠁⠈⠐⠠⢀⡀⠄⠂":
            yield c


def page_lines(lines):
    """
    View the given list of Unicode lines in a pager (e.g. `less`).
    """
    lines = list(lines)
    line_width = max([len(ansi_clean(line)) for line in lines])
    if (
        TTY and (
            line_width > get_terminal_size().columns or
            len(lines) > get_terminal_size().lines
        )
    ):
        write_to_stream(STDOUT_WRITER, SHOW_CURSOR)
        env = environ.copy()
        env["LESS"] = env.get("LESS", "") + " -R"
        pager = Popen(
            [environ.get("PAGER", "/usr/bin/less")],
            env=env,
            stdin=PIPE,
        )
        with suppress(BrokenPipeError):
            pager.stdin.write("\n".join(lines).encode('utf-8'))
            pager.stdin.close()
        pager.communicate()
        write_to_stream(STDOUT_WRITER, HIDE_CURSOR)
    else:
        for line in lines:
            io.stdout(line)


def write_to_stream(stream, msg, flush=True):
    with suppress(BrokenPipeError):
        if TTY:
            stream.write(msg)
        else:
            stream.write(ansi_clean(msg))

        if flush:
            stream.flush()


class DrainableStdin:
    def get_input(self):
        while True:
            if QUIT_EVENT.is_set():
                return None
            if select([sys.stdin], [], [], 0.1)[0]:
                return sys.stdin.readline().strip()

    def drain(self):
        if sys.stdin.isatty():
            termios.tcflush(sys.stdin, termios.TCIFLUSH)


class JobManager:
    def __init__(self):
        self._jobs = []

    def add(self, msg):
        job_id = (time(), msg)
        self._jobs.append(job_id)
        return job_id

    def remove(self, job_id):
        self._jobs.remove(job_id)

    @property
    def current_job(self):
        try:
            job_start, job_msg = self._jobs[-1]
        except IndexError:
            return None
        current_time = time()
        if current_time - job_start > 3.0:
            # If the latest job is taking a long time, start rotating
            # the displayed job every 3s. That way, users can see all
            # long-running jobs currently in progress.
            index = int(current_time / 3.0) % len(self._jobs)
            job_start, job_msg = self._jobs[index]

        elapsed = current_time - job_start
        if elapsed > 10.0:
            job_msg += " ({})".format(format_duration(timedelta(seconds=elapsed)))

        return job_msg

    @property
    def messages(self):
        return [job_msg for job_start, job_msg in self._jobs]


class IOManager:
    """
    Threadsafe singleton class that handles all IO.
    """
    def __init__(self):
        self._active = False
        self.debug_log_file = None
        self.debug_mode = False
        self.jobs = JobManager()
        self.lock = Lock()
        self.progress = 0
        self.progress_start = None
        self.progress_total = 0
        self._spinner = spinner()
        self._last_spinner_character = next(self._spinner)
        self._last_spinner_update = 0
        self._signal_handler_thread = None
        self._child_pids = []
        self._status_line_present = False
        self._waiting_for_input = False

    def __enter__(self):
        self.activate()
        # return self, so users could also do `with IOManager() as io:`
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.deactivate()

    def activate(self):
        if self._active:
            return

        self._active = True
        if 'BW_DEBUG_LOG_DIR' in environ:
            self.debug_log_file = open(join(
                environ['BW_DEBUG_LOG_DIR'],
                "{}_{}.log".format(
                    datetime.now().strftime("%Y-%m-%d_%H-%M-%S"),
                    getpid(),
                ),
            ), 'a')
        self._signal_handler_thread = Thread(
            target=self._signal_handler_thread_body,
        )
        # daemon mode is required because we need to keep the thread
        # around until the end of a soft shutdown to wait for a hard
        # shutdown signal, but don't have a feasible way of stopping
        # the thread once the soft shutdown has completed
        self._signal_handler_thread.daemon = True
        self._signal_handler_thread.start()
        self._original_sigint_handler = signal(SIGINT, sigint_handler)
        self._original_sigquit_handler = signal(SIGQUIT, sigquit_handler)
        faulthandler.enable()
        if TTY:
            write_to_stream(STDOUT_WRITER, HIDE_CURSOR)

    def ask(self, question, default, epilogue=None, input_handler=DrainableStdin()):
        assert self._active
        answers = _("[Y/n]") if default else _("[y/N]")
        question = question + " " + answers + " "
        self._waiting_for_input = True
        with self.lock:
            if QUIT_EVENT.is_set():
                sys.exit(0)
            self._clear_last_job(flush=False)
            while True:
                write_to_stream(STDOUT_WRITER, "\a" + question + SHOW_CURSOR)

                input_handler.drain()
                answer = input_handler.get_input()
                if answer is None:
                    if epilogue:
                        write_to_stream(STDOUT_WRITER, "\n" + epilogue + "\n")
                    QUIT_EVENT.set()
                    sys.exit(0)
                elif answer.lower() in (_("y"), _("yes")) or (
                    not answer and default
                ):
                    answer = True
                    break
                elif answer.lower() in (_("n"), _("no")) or (
                    not answer and not default
                ):
                    answer = False
                    break
                write_to_stream(
                    STDOUT_WRITER,
                    _("Please answer with 'y(es)' or 'n(o)'.\n"),
                )
            if epilogue:
                write_to_stream(STDOUT_WRITER, epilogue + "\n")
            write_to_stream(STDOUT_WRITER, HIDE_CURSOR)
        self._waiting_for_input = False
        return answer

    def deactivate(self):
        if not self._active:
            return

        self._active = False
        if TTY:
            write_to_stream(STDOUT_WRITER, SHOW_CURSOR)
        signal(SIGINT, self._original_sigint_handler)
        signal(SIGQUIT, self._original_sigint_handler)
        faulthandler.disable()
        self._signal_handler_thread.join()
        if self.debug_log_file:
            self.debug_log_file.close()

    @clear_formatting
    @add_debug_indicator
    @capture_for_debug_logfile
    @add_debug_timestamp
    def debug(self, msg, append_newline=True):
        if self.debug_mode:
            with self.lock:
                self._write(msg, append_newline=append_newline)

    def job_add(self, msg):
        if not self._active:
            return
        with self.lock:
            self._clear_last_job(flush=False)
            job_id = self.jobs.add(msg)
            self._write_current_job()
            return job_id

    def job_del(self, job_id):
        if not self._active:
            return
        with self.lock:
            self._clear_last_job(flush=False)
            self.jobs.remove(job_id)
            self._write_current_job()

    def progress_advance(self, increment=1):
        with self.lock:
            self.progress += increment

    def progress_increase_total(self, increment=1):
        with self.lock:
            self.progress_total += increment

    def progress_set_total(self, total):
        self.progress = 0
        self.progress_start = datetime.utcnow()
        self.progress_total = total

    def progress_show(self):
        if INFO_EVENT.is_set():
            INFO_EVENT.clear()
            table = []
            if self.jobs.messages:
                table.append([bold(_("Running jobs")), self.jobs.messages[0].strip()])
                for job in self.jobs.messages[1:]:
                    table.append(["", job.strip()])
            try:
                progress = (self.progress / float(self.progress_total))
                elapsed = datetime.utcnow() - self.progress_start
                remaining = elapsed / progress - elapsed
            except ZeroDivisionError:
                pass
            else:
                if table:
                    table.append(ROW_SEPARATOR)
                table.extend([
                    [bold(_("Progress")), "{:.1f}%".format(progress * 100)],
                    ROW_SEPARATOR,
                    [bold(_("Elapsed")), format_duration(elapsed)],
                    ROW_SEPARATOR,
                    [
                        bold(_("Remaining")),
                        _("{} (estimate based on progress)").format(format_duration(remaining))
                    ],
                ])
            output = blue("i") + "\n"
            if table:
                for line in render_table(table):
                    output += ("{x} {line}\n".format(x=blue("i"), line=line))
            else:
                output += _("{x}  No progress info available at this time.\n").format(x=blue("i"))
            io.stderr(output + blue("i"))

    @clear_formatting
    @capture_for_debug_logfile
    @add_debug_timestamp
    def stderr(self, msg, append_newline=True):
        with self.lock:
            self._write(msg, append_newline=append_newline, err=True)

    @clear_formatting
    @capture_for_debug_logfile
    @add_debug_timestamp
    def stdout(self, msg, append_newline=True):
        with self.lock:
            self._write(msg, append_newline=append_newline)

    @contextmanager
    def job(self, job_text):
        job_id = self.job_add(job_text)
        try:
            yield
        finally:
            self.job_del(job_id)

    def job_wrapper(self, job_text):
        def outer_wrapper(wrapped_function):
            @wraps(wrapped_function)
            def inner_wrapper(*args, **kwargs):
                with self.job(job_text.format(*args, **kwargs)):
                    return wrapped_function(*args, **kwargs)
            return inner_wrapper
        return outer_wrapper

    def _clear_last_job(self, flush=True):
        if self._status_line_present and TTY:
            # Some terminals respond very quickly to \e[K: They
            # immediately clear the line and redraw their screen right
            # away, which means an empty line is visible for a brief
            # moment. It takes some (short amount of) time until we
            # write the next line. This means we alternate between
            # "empty line" and "line with text", which means:
            # Flickering.
            #
            # We can avoid this to some degree by not flushing stdout
            # after writing \e[K.
            #
            # We can only do this if we know for certain that we will
            # write more data soon. For example, this is the case when
            # _clear_last_job() is followed by _write_current_job().
            #
            # It is *not* possible to always omit the flush. Sometimes,
            # _clear_last_job() is called before _write(..., err=True),
            # meaning the first call writes to stdout and the second one
            # to stderr. This results in garbled output because the \e[K
            # is still lingering in the buffer for stdout.
            #
            # Note to future readers: If this causes issues and it's too
            # hard to understand, then simply remove the 'flush' flag
            # and go back to always flushing the buffers. This is mostly
            # a cosmetic issue and it's better to have some flickering
            # than completely garbled output.
            write_to_stream(STDOUT_WRITER, "\r\033[K", flush=flush)
            self._status_line_present = False

    def _signal_handler_thread_body(self):
        while self._active:
            self.progress_show()
            if not self._waiting_for_input:  # do not block and ignore SIGINT while .ask()ing
                with self.lock:
                    self._clear_last_job(flush=False)
                    self._write_current_job()
            if QUIT_EVENT.is_set():
                if SHUTDOWN_EVENT_HARD.wait(0.1):
                    self.stderr(_("{x} {signal}  cleanup interrupted, exiting...").format(
                        signal=bold(_("SIGINT")),
                        x=blue("i"),
                    ))
                    for ssh_pid in self._child_pids:
                        self.debug(_("killing SSH session with PID {pid}").format(pid=ssh_pid))
                        with suppress(ProcessLookupError):
                            kill(ssh_pid, SIGTERM)
                    self._clear_last_job()
                    if TTY:
                        write_to_stream(STDOUT_WRITER, SHOW_CURSOR)
                    _exit(130)  # https://tldp.org/LDP/abs/html/exitcodes.html
            else:
                if SHUTDOWN_EVENT_SOFT.wait(0.1):
                    QUIT_EVENT.set()
                    self.stderr(_(
                        "{x} {signal}  canceling pending tasks... "
                        "(hit CTRL+C again for immediate dirty exit)"
                    ).format(
                        signal=bold(_("SIGINT")),
                        x=blue("i"),
                    ))

    def _spinner_character(self):
        if time() - self._last_spinner_update > 0.2:
            self._last_spinner_update = time()
            self._last_spinner_character = next(self._spinner)
        return self._last_spinner_character

    def _write(self, msg, append_newline=True, err=False):
        if not self._active:
            return
        self._clear_last_job()
        if msg is not None:
            if append_newline:
                msg += "\n"
            write_to_stream(STDERR_WRITER if err else STDOUT_WRITER, msg)
        self._write_current_job()

    def _write_current_job(self):
        current_job = self.jobs.current_job
        if TTY:
            if current_job:
                line = "{} ".format(blue(self._spinner_character()))
                try:
                    progress = (self.progress / float(self.progress_total))
                except ZeroDivisionError:
                    pass
                else:
                    progress_text = "{:.1f}%  ".format(progress * 100)
                    line += bold(progress_text)
                line += current_job
                write_to_stream(
                    STDOUT_WRITER,
                    trim_visible_len_to(line, get_terminal_size().columns),
                )
                self._status_line_present = True
            else:
                # If there are jobs to be displayed, the call above to
                # write_to_stream() will flush stdout. But if there are
                # none, then a \e[K might still be lingering in the
                # buffer. See comment in _clear_last_job().
                STDOUT_WRITER.flush()


io = IOManager()



from functools import wraps
from os import environ
from sys import exit, stderr, stdout
from traceback import print_exc

from ..exceptions import NoSuchGroup, NoSuchItem, NoSuchNode, NoSuchTarget
from .text import mark_for_translation as _, red
from .ui import io, QUIT_EVENT


def exit_on_keyboardinterrupt(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except KeyboardInterrupt:
            exit(130)  # https://tldp.org/LDP/abs/html/exitcodes.html
    return wrapper


def suppress_broken_pipe_msg(f):
    """
    Oh boy.

    CPython does funny things with SIGPIPE. By default, it is caught and
    raised as a BrokenPipeError. When do we get a SIGPIPE? Most commonly
    when piping into head:

        bw nodes | head -n 1

    head will exit after receiving the first line, causing the kernel to
    send SIGPIPE to our process. Since in most cases, we can't just quit
    early, we simply ignore BrokenPipeError in utils.ui.write_to_stream.

    Unfortunately, Python will still print a message:

        Exception ignored in: <_io.TextIOWrapper name='<stdout>'
                               mode='w' encoding='UTF-8'>
        BrokenPipeError: [Errno 32] Broken pipe

    See also http://bugs.python.org/issue11380. The crazy try/finally
    construct below is taken from there and I quote:

        This will:
         - capture any exceptions *you've* raised as the context for the
           errors raised in this handler
         - expose any exceptions generated during this thing itself
         - prevent the interpreter dying during shutdown in
           flush_std_files by closing the files (you can't easily wipe
           out the pending writes that have failed)

    CAVEAT: There is a seamingly easier method floating around on the
    net (http://stackoverflow.com/a/16865106) that restores the default
    behavior for SIGPIPE (i.e. not turning it into a BrokenPipeError):

        from signal import signal, SIGPIPE, SIG_DFL
        signal(SIGPIPE,SIG_DFL)

    This worked fine for a while but broke when using
    multiprocessing.Manager() to share the list of jobs in utils.ui
    between processes. When the main process terminated, it quit with
    return code 141 (indicating a broken pipe), and the background
    process used for the manager continued to hang around indefinitely.
    Bonus fun: This was observed only on Ubuntu Trusty (14.04).
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception:
            print_exc()
            exit(1)
        finally:
            try:
                stdout.flush()
            finally:
                try:
                    stdout.close()
                finally:
                    try:
                        stderr.flush()
                    finally:
                        stderr.close()
    return wrapper


def count_items(nodes):
    count = 0
    for node in nodes:
        if QUIT_EVENT.is_set():
            return 0
        count += len(node.items)
    return count


def get_group(repo, group_name):
    try:
        return repo.get_group(group_name)
    except NoSuchGroup:
        io.stderr(_("{x} No such group: {group}").format(
            group=group_name,
            x=red("!!!"),
        ))
        exit(1)


def get_item(node, item_id):
    try:
        return node.get_item(item_id)
    except NoSuchItem:
        io.stderr(_("{x} No such item on node '{node}': {item}").format(
            item=item_id,
            node=node.name,
            x=red("!!!"),
        ))
        exit(1)


def get_node(repo, node_name):
    try:
        return repo.get_node(node_name)
    except NoSuchNode:
        io.stderr(_("{x} No such node: {node}").format(
            node=node_name,
            x=red("!!!"),
        ))
        exit(1)



DEFAULT_item_workers = int(environ.get("BW_ITEM_WORKERS", "4"))
DEFAULT_node_workers = int(environ.get("BW_NODE_WORKERS", "4"))
DEFAULT_softlock_expiry = environ.get("BW_SOFTLOCK_EXPIRY", "8h")

HELP_get_target_nodes = _("""expression to select target nodes:

my_node            # to select a single node
my_group           # all nodes in this group
bundle:my_bundle   # all nodes with this bundle
!bundle:my_bundle  # all nodes without this bundle
!group:my_group    # all nodes not in this group
"lambda:node.metadata_get('foo/magic', 47) < 3"
                   # all nodes whose metadata.get("foo/magic") is less than three
""")
HELP_item_workers = _("number of items to apply simultaneously on each node (defaults to {})").format(DEFAULT_item_workers)
HELP_node_workers = _("number of nodes to apply to simultaneously (defaults to {})").format(DEFAULT_node_workers)
HELP_softlock_expiry = _("how long before the lock is ignored and removed automatically (defaults to \"{}\")").format(DEFAULT_softlock_expiry)


def get_target_nodes(repo, target_strings, node_workers=None):
    if not node_workers:
        node_workers = DEFAULT_node_workers

    try:
        nodes_matching = repo.nodes_matching(target_strings, node_workers)
        if not nodes_matching:
            io.stderr(_("{x} Input did not match any nodes").format(
                x=red("!!!"),
            ))
            exit(1)
    except NoSuchTarget as e:
        io.stderr("{x} Target string {name} does match neither bundle, nor group, node or lambda.".format(
            x=red("!!!"),
            name = e,
            ))
        exit(1)

    return nodes_matching


def verify_autoskip_selectors(nodes, selectors_given):
    if not selectors_given:
        return set()

    selectors_used = {
        selector: False
        for selector in selectors_given
    }
    for node in nodes:
        for selector in selectors_given:
            if node.covered_by_autoskip_selector([selector]):
                selectors_used[selector] = True

        for item in node.items:
            if QUIT_EVENT.is_set():
                return 0

            for selector in selectors_given:
                if item.covered_by_autoskip_selector([selector]):
                    selectors_used[selector] = True

    return {
        selector
        for selector, matches in selectors_used.items()
        if not matches
    }
